import { __awaiter, __generator } from "tslib";
import { from, race, timer, throwError, Subject, noop, Observable } from 'rxjs';
import { flatMap, bufferCount, take, filter, tap } from 'rxjs/operators';
import { rootInjector } from '@asuka/di';
import { SSRSymbol } from './constants';
import { SKIP_SYMBOL } from './ssr-effect';
import { TERMINATE_ACTION } from './terminate';
import { SSRStateCacheInstance } from './ssr-states';
import { oneShotCache } from './ssr-oneshot-cache';
var skipFn = function () { return SKIP_SYMBOL; };
/**
 * Run all @SSREffect decorated effects of given modules and extract latest states.
 * `cleanup` function returned must be called before end of responding
 *
 * @param ctx request context, which will be passed to payloadGetter in SSREffect decorator param
 * @param modules used ayanami modules
 * @param uuid the same uuid would reuse the same state which was created before
 * @param timeout seconds to wait before all effects stream out TERMINATE_ACTION
 * @returns object contains ayanami state and cleanup function
 */
export var runSSREffects = function (ctx, modules, sharedCtx, timeout) {
    if (timeout === void 0) { timeout = 3; }
    var stateToSerialize = {};
    return modules.length === 0
        ? Promise.resolve(stateToSerialize)
        : race(from(modules).pipe(flatMap(function (constructor) {
            return new Observable(function (observer) {
                var cleanup = noop;
                var metas = Reflect.getMetadata(SSRSymbol, constructor.prototype) || [];
                var ayanamiState;
                var moduleName;
                var middleware = function (effect$) {
                    return effect$.pipe(tap({
                        error: function (e) {
                            observer.error(e);
                        },
                    }));
                };
                if (sharedCtx) {
                    if (SSRStateCacheInstance.has(sharedCtx, constructor)) {
                        ayanamiState = SSRStateCacheInstance.get(sharedCtx, constructor);
                        moduleName = constructor.prototype.moduleName;
                    }
                    else {
                        var ayanamiInstance = rootInjector.resolveAndInstantiate(constructor);
                        moduleName = ayanamiInstance.moduleName;
                        ayanamiState = ayanamiInstance.createState(middleware);
                        SSRStateCacheInstance.set(sharedCtx, constructor, ayanamiState);
                    }
                }
                else {
                    var ayanamiInstance = rootInjector.resolveAndInstantiate(constructor);
                    moduleName = ayanamiInstance.moduleName;
                    ayanamiState = ayanamiInstance.createState(middleware);
                    oneShotCache.store(ctx, constructor, ayanamiState);
                }
                var effectsCount = metas.length;
                var disposeFn = noop;
                cleanup = sharedCtx
                    ? function () { return disposeFn(); }
                    : function () {
                        ayanamiState.unsubscribe();
                    };
                function runEffects() {
                    return __awaiter(this, void 0, void 0, function () {
                        var action$_1, state;
                        var _this = this;
                        return __generator(this, function (_a) {
                            switch (_a.label) {
                                case 0: return [4 /*yield*/, Promise.all(metas.map(function (meta) { return __awaiter(_this, void 0, void 0, function () {
                                        var param;
                                        return __generator(this, function (_a) {
                                            switch (_a.label) {
                                                case 0:
                                                    if (!meta.middleware) return [3 /*break*/, 2];
                                                    return [4 /*yield*/, meta.middleware(ctx, skipFn)];
                                                case 1:
                                                    param = _a.sent();
                                                    if (param !== SKIP_SYMBOL) {
                                                        ayanamiState.dispatch({
                                                            type: meta.action,
                                                            payload: param,
                                                            state: ayanamiState,
                                                        });
                                                    }
                                                    else {
                                                        effectsCount -= 1;
                                                    }
                                                    return [3 /*break*/, 3];
                                                case 2:
                                                    ayanamiState.dispatch({
                                                        type: meta.action,
                                                        payload: undefined,
                                                        state: ayanamiState,
                                                    });
                                                    _a.label = 3;
                                                case 3: return [2 /*return*/];
                                            }
                                        });
                                    }); }))];
                                case 1:
                                    _a.sent();
                                    if (!(effectsCount > 0)) return [3 /*break*/, 3];
                                    action$_1 = new Subject();
                                    disposeFn = ayanamiState.subscribeAction(function (action) {
                                        action$_1.next(action);
                                    });
                                    return [4 /*yield*/, action$_1
                                            .pipe(filter(function (act) { return act.type === TERMINATE_ACTION.type; }), bufferCount(effectsCount), take(1))
                                            .toPromise()];
                                case 2:
                                    _a.sent();
                                    state = ayanamiState.getState();
                                    stateToSerialize[moduleName] = state;
                                    _a.label = 3;
                                case 3: return [2 /*return*/];
                            }
                        });
                    });
                }
                runEffects()
                    .then(function () {
                    observer.next(stateToSerialize);
                    observer.complete();
                })
                    .catch(function (e) {
                    observer.error(e);
                });
                return cleanup;
            });
        })), timer(timeout * 1000).pipe(flatMap(function () { return throwError(new Error('Terminate timeout')); }))).toPromise();
};
