import { __decorate, __metadata } from "tslib";
import 'reflect-metadata';
import React, { useEffect, useCallback } from 'react';
import { useAyanami, useAyanamiState, useAyanamiDispatchers } from '../hooks';
import { Ayanami, Effect, ImmerReducer, Module, Reducer } from '../core';
import { Observable } from 'rxjs';
import { map, withLatestFrom, delay, skipWhile, combineLatest, take } from 'rxjs/operators';
import { act, create } from 'react-test-renderer';
import { rootInjector, InjectableContext } from '@asuka/di';
import Sinon from 'sinon';
const ASYNC_DELAY_TIME = 1000;
let GlobalModule = class GlobalModule extends Ayanami {
    constructor() {
        super(...arguments);
        this.defaultState = {
            user: '',
        };
    }
    setUser(state, payload) {
        state.user = payload;
    }
    fetchUser(payload$) {
        return payload$.pipe(delay(ASYNC_DELAY_TIME), map(() => this.getActions().setUser('global')));
    }
};
__decorate([
    ImmerReducer(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object, String]),
    __metadata("design:returntype", void 0)
], GlobalModule.prototype, "setUser", null);
__decorate([
    Effect(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Observable]),
    __metadata("design:returntype", Observable)
], GlobalModule.prototype, "fetchUser", null);
GlobalModule = __decorate([
    Module('Global')
], GlobalModule);
let CountModule = class CountModule extends Ayanami {
    constructor(globalModule) {
        super();
        this.globalModule = globalModule;
        this.defaultState = {
            count: 0,
        };
    }
    setCount(state, payload) {
        state.count = payload;
    }
    addCount(payload$) {
        return payload$.pipe(withLatestFrom(this.state$), map(([, state]) => this.getActions().setCount(state.count + 1)));
    }
    addCountAsync(payload$) {
        return payload$.pipe(delay(ASYNC_DELAY_TIME), withLatestFrom(this.state$), map(([, state]) => this.getActions().setCount(state.count + 1)));
    }
    waitUserAddCount(payload$) {
        return payload$.pipe(combineLatest(this.globalModule.state$.pipe(skipWhile((v) => !v.user), take(1)), this.state$.pipe(take(1))), map(([, , state]) => this.getActions().setCount(state.count + 1)));
    }
};
__decorate([
    ImmerReducer(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object, Number]),
    __metadata("design:returntype", void 0)
], CountModule.prototype, "setCount", null);
__decorate([
    Effect(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Observable]),
    __metadata("design:returntype", Observable)
], CountModule.prototype, "addCount", null);
__decorate([
    Effect(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Observable]),
    __metadata("design:returntype", Observable)
], CountModule.prototype, "addCountAsync", null);
__decorate([
    Effect(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Observable]),
    __metadata("design:returntype", Observable)
], CountModule.prototype, "waitUserAddCount", null);
CountModule = __decorate([
    Module('CountModule'),
    __metadata("design:paramtypes", [GlobalModule])
], CountModule);
describe('Hooks specs: useAyanami', () => {
    const SimpleComponent = () => {
        const [state, actions] = useAyanami(CountModule);
        const addCount = useCallback(() => {
            actions.addCount();
        }, []);
        const addCountAsync = useCallback(() => {
            actions.addCountAsync();
        }, []);
        return (React.createElement("div", null,
            React.createElement("span", null, state.count),
            React.createElement("button", { id: "add-count", onClick: addCount }, "+"),
            React.createElement("button", { id: "add-count-async", onClick: addCountAsync }, "+"),
            React.createElement("button", { id: "set-count", onClick: () => actions.setCount(10) }, "set")));
    };
    let testRenderer;
    let getCount;
    let timer;
    beforeEach(() => {
        timer = Sinon.useFakeTimers();
        testRenderer = create(React.createElement(InjectableContext, null,
            React.createElement(SimpleComponent, null)));
        getCount = () => parseInt(testRenderer.root.find((instance) => instance.type === 'span').children[0]);
    });
    afterEach(() => {
        timer.restore();
        rootInjector.reset();
        rootInjector.addProviders([GlobalModule, CountModule]);
    });
    it('should render correct state', () => {
        expect(getCount()).toBe(0);
    });
    it('should dispatch action', () => {
        act(() => {
            testRenderer.root.find((instance) => instance.props.id === 'set-count').props.onClick();
        });
        expect(getCount()).toBe(10);
    });
    it('should dispatch effect', () => {
        expect(getCount()).toBe(0);
        act(() => {
            testRenderer.root.find((instance) => instance.props.id === 'add-count').props.onClick();
        });
        expect(getCount()).toBe(1);
    });
    it('should dispatch async effect', () => {
        act(() => {
            testRenderer.root.find((instance) => instance.props.id === 'add-count-async').props.onClick();
            timer.tick(ASYNC_DELAY_TIME);
        });
        expect(getCount()).toBe(1);
    });
    it('should only render once when update state right during rendering', () => {
        const spyFn = jest.fn();
        const TestComponent = () => {
            const [state, actions] = useAyanami(CountModule);
            if (state.count === 2) {
                actions.addCount();
            }
            useEffect(() => {
                spyFn(state.count);
            }, [state.count]);
            return (React.createElement("div", null,
                React.createElement("span", null, state.count),
                React.createElement("button", { id: "set-count", onClick: () => actions.setCount(2) }, "set")));
        };
        const renderer = create(React.createElement(TestComponent, null));
        // https://github.com/facebook/react/issues/14050 to trigger useEffect manually
        renderer.update(React.createElement(TestComponent, null));
        expect(spyFn.mock.calls).toEqual([[0]]);
        act(() => {
            renderer.root.findByType('button').props.onClick();
        });
        expect(spyFn.mock.calls).toEqual([[0], [3]]);
    });
    it('should be able to use state$ from the other module', () => {
        const TestComponent = () => {
            const [state, actions] = useAyanami(CountModule);
            const globalActions = useAyanamiDispatchers(GlobalModule);
            useEffect(() => {
                actions.waitUserAddCount();
                globalActions.fetchUser();
            }, []);
            return React.createElement("span", null, state.count);
        };
        const renderer = create(React.createElement(TestComponent, null));
        act(() => {
            renderer.update(React.createElement(TestComponent, null));
        });
        act(() => {
            timer.tick(ASYNC_DELAY_TIME);
            renderer.update(React.createElement(TestComponent, null));
        });
        expect(getCount()).toBe(1);
    });
    it('should be able to catch reducer error', () => {
        const error = new TypeError('whatever');
        let ReducerErrorModule = class ReducerErrorModule extends Ayanami {
            constructor() {
                super(...arguments);
                this.defaultState = {};
            }
            throw() {
                throw error;
            }
        };
        __decorate([
            Reducer(),
            __metadata("design:type", Function),
            __metadata("design:paramtypes", []),
            __metadata("design:returntype", void 0)
        ], ReducerErrorModule.prototype, "throw", null);
        ReducerErrorModule = __decorate([
            Module('ReducerError')
        ], ReducerErrorModule);
        const spy = Sinon.spy();
        const Component = () => {
            const actions = useAyanamiDispatchers(ReducerErrorModule);
            try {
                actions.throw();
            }
            catch (e) {
                spy(e);
            }
            return React.createElement("div", null);
        };
        create(React.createElement(InjectableContext, null,
            React.createElement(Component, null)));
        expect(spy.args).toStrictEqual([[error]]);
    });
});
describe('Hooks spec: useAyanami with config', () => {
    const Component = () => {
        const [count, actions] = useAyanami(CountModule, {
            selector: (state) => state.count,
            mutateStateOnFirstRendering: (state) => {
                state.count = 2;
            },
        });
        const addCount = useCallback(() => {
            actions.addCount();
        }, []);
        return (React.createElement("div", null,
            React.createElement("span", null, count),
            React.createElement("button", { id: "add-count", onClick: addCount }, "+")));
    };
    let testRenderer;
    let getCount;
    beforeEach(() => {
        testRenderer = create(React.createElement(Component, null));
        getCount = () => parseInt(testRenderer.root.find((instance) => instance.type === 'span').children[0]);
    });
    afterEach(() => {
        rootInjector.reset();
        rootInjector.addProviders([GlobalModule, CountModule]);
    });
    it('should return selected state from store', () => {
        expect(getCount()).toBe(2);
    });
    it('should only mutate state on first rendering', () => {
        expect(getCount()).toBe(2);
        act(() => {
            testRenderer.root.findByType('button').props.onClick();
        });
        expect(getCount()).toBe(3);
    });
});
describe('Hooks spec: useAyanamiState', () => {
    const Component = () => {
        const count = useAyanamiState(CountModule, {
            selector: (state) => state.count,
        });
        return (React.createElement("div", null,
            React.createElement("span", null, count)));
    };
    let testRenderer;
    let getCount;
    beforeEach(() => {
        testRenderer = create(React.createElement(Component, null));
        getCount = () => parseInt(testRenderer.root.find((instance) => instance.type === 'span').children[0]);
    });
    afterEach(() => {
        rootInjector.reset();
        rootInjector.addProviders([GlobalModule, CountModule]);
    });
    it('should return selected state from store', () => {
        expect(getCount()).toBe(0);
    });
});
describe('Hooks spec: useAyanamiDispatchers', () => {
    const Component = () => {
        const state = useAyanamiState(CountModule);
        const actions = useAyanamiDispatchers(CountModule);
        const addCount = useCallback(() => {
            actions.addCount();
        }, []);
        return (React.createElement("div", null,
            React.createElement("span", null, state.count),
            React.createElement("button", { id: "add-count", onClick: addCount }, "+")));
    };
    let testRenderer;
    let getCount;
    beforeEach(() => {
        testRenderer = create(React.createElement(Component, null));
        getCount = () => parseInt(testRenderer.root.find((instance) => instance.type === 'span').children[0]);
    });
    afterEach(() => {
        rootInjector.reset();
        rootInjector.addProviders([GlobalModule, CountModule]);
    });
    it('should return selected state from store', () => {
        expect(getCount()).toBe(0);
    });
    it('should dispatch effect', () => {
        act(() => {
            testRenderer.root.findByType('button').props.onClick();
        });
        expect(getCount()).toBe(1);
    });
});
