import { createState } from '../state';
import { identity } from 'rxjs';
import { filter, map, delay } from 'rxjs/operators';
import * as Sinon from 'sinon';
describe('state specs', () => {
    const UPDATE_FOO = 'update-foo';
    const UPDATE_BAR = 'update-bar';
    const ASYNC_UPDATE_FOO = 'async-update-foo';
    const ASYNC_UPDATE_BAR = 'async-update-bar';
    const mockReducer = (prevState, action) => {
        if (action.type === UPDATE_FOO) {
            return { ...prevState, foo: action.payload };
        }
        else if (action.type === UPDATE_BAR) {
            return { ...prevState, bar: action.payload };
        }
        return prevState;
    };
    let state;
    const mockEpic = (action$) => action$.pipe(filter(({ type }) => type === ASYNC_UPDATE_FOO), map(() => ({ type: 'noop', payload: null, state })));
    describe('stateCreator', () => {
        const defaultState = {
            foo: '1',
            bar: null,
        };
        const { stateCreator } = createState(mockReducer, identity);
        it('should be able to create state', () => {
            expect(stateCreator(defaultState).getState()).toBe(defaultState);
        });
        it('should be able to create state with new defaultState', () => {
            const newDefaultState = { ...defaultState };
            expect(stateCreator(newDefaultState).getState()).toBe(newDefaultState);
        });
    });
    describe('state', () => {
        const defaultState = {
            foo: '1',
            bar: null,
        };
        it('should be able to change state via dispatch action', () => {
            const { stateCreator } = createState(mockReducer, mockEpic);
            const state = stateCreator(defaultState);
            state.dispatch({ type: UPDATE_FOO, payload: '2', state });
            expect(state.getState().foo).toBe('2');
        });
        it('should do nothing when action type is not matched any reducer', () => {
            const { stateCreator } = createState(mockReducer, mockEpic);
            const state = stateCreator(defaultState);
            state.dispatch({ type: '__NOOP__', payload: '2', state });
            expect(state.getState().foo).toBe(defaultState.foo);
        });
        it('should run epic', () => {
            const { stateCreator } = createState(mockReducer, mockEpic);
            const state = stateCreator(defaultState);
            state.dispatch({ type: UPDATE_FOO, payload: '2', state });
            expect(state.getState().foo).toBe('2');
        });
        describe('effect', () => {
            let stateCreator;
            let timer;
            const bar = 100;
            const delayTime = 300;
            beforeEach(() => {
                const { stateCreator: _stateCreator } = createState(mockReducer, (action$) => action$.pipe(filter(({ type }) => type === ASYNC_UPDATE_BAR), delay(delayTime), map(() => ({
                    type: UPDATE_BAR,
                    payload: bar,
                    state,
                }))));
                stateCreator = _stateCreator;
                timer = Sinon.useFakeTimers();
            });
            afterEach(() => {
                timer.restore();
            });
            it('should be able to change state by effect', () => {
                state = stateCreator(defaultState);
                state.dispatch({
                    type: ASYNC_UPDATE_BAR,
                    payload: null,
                    state,
                });
                timer.tick(delayTime);
                expect(state.getState().bar).toBe(bar);
            });
        });
    });
});
