import { merge, identity } from 'rxjs';
import { map, filter, publish, refCount, skip } from 'rxjs/operators';
import produce from 'immer';
import { createState } from './state';
import { EFFECT_DECORATOR_SYMBOL, REDUCER_DECORATOR_SYMBOL, IMMER_REDUCER_DECORATOR_SYMBOL, DEFINE_ACTION_DECORATOR_SYMBOL, } from './symbols';
import { GLOBAL_KEY, ACTION_TO_SKIP_KEY, SSR_LOADED_KEY } from '../ssr/constants';
import { logStateAction, INIT_ACTION_TYPE } from '../redux-devtools-extension';
const _globalThis = 
/* istanbul ignore next */ typeof globalThis === 'undefined'
    ? /* istanbul ignore next */ typeof window === 'undefined'
        ? global
        : window
    : globalThis;
// cast to string to match the shape of action
const NOOP_ACTION_TYPE = Symbol('NOOP_ACTION');
const effectNameSymbol = Symbol('effect-name');
export class Ayanami {
    constructor() {
        // @internal
        this._actionKeys = [];
        // @internal
        this.state = null;
        this.effect = this.combineEffects();
        this.reducer = this.combineReducers();
        const { stateCreator, action$, state$ } = createState(this.reducer, this.effect);
        this.stateCreator = stateCreator;
        this.action$ = action$;
        this.state$ = state$;
        this.combineDefineActions();
        this._actions = this._actionKeys.reduce((acc, key) => {
            acc[key] = (payload) => {
                const action = {
                    type: key,
                    payload,
                };
                Object.defineProperty(action, 'state', {
                    value: this.state,
                    enumerable: false,
                    configurable: false,
                    writable: false,
                });
                return action;
            };
            return acc;
        }, Object.create(null));
        this._actionStreams = this._actionKeys.reduce((acc, key) => {
            acc[key] = this.action$.pipe(filter(({ type }) => type === key), map(({ payload }) => payload));
            return acc;
        }, {});
    }
    createState(middleware = identity) {
        if (this.state)
            return this.state;
        const ssrCache = _globalThis[Symbol.for(Symbol.keyFor(GLOBAL_KEY))];
        let loadFromSSR = false;
        let preloadState;
        if (ssrCache && ssrCache[this.moduleName]) {
            preloadState = ssrCache[this.moduleName];
            loadFromSSR = true;
        }
        this.state = this.stateCreator(preloadState ?? this.defaultState, middleware, loadFromSSR);
        Reflect.defineMetadata(SSR_LOADED_KEY, loadFromSSR, this.state);
        if (process.env.NODE_ENV !== 'production') {
            Object.defineProperty(this.state, 'name', {
                value: this.moduleName,
                configurable: false,
                enumerable: false,
                writable: false,
            });
            logStateAction({
                type: INIT_ACTION_TYPE,
                state: this.state,
                payload: null,
            });
        }
        return this.state;
    }
    getActions() {
        return this._actions;
    }
    getAction$() {
        return this._actionStreams;
    }
    createNoopAction() {
        return {
            type: NOOP_ACTION_TYPE,
            payload: null,
            state: this.state,
        };
    }
    combineEffects() {
        const effectKeys = Reflect.getMetadata(EFFECT_DECORATOR_SYMBOL, this.constructor) || [];
        this._actionKeys.push(...effectKeys);
        const effects = effectKeys.map((property) => {
            const effect = this[property].bind(this);
            Object.defineProperty(effect, effectNameSymbol, {
                value: property,
                enumerable: false,
                configurable: false,
                writable: false,
            });
            return effect;
        });
        return (action$, loadFromSSR) => {
            const actionsToSkip = Reflect.getMetadata(ACTION_TO_SKIP_KEY, this.constructor.prototype);
            return merge(...effects.map((effect) => {
                const effectActionType = effect[effectNameSymbol];
                const skipCount = loadFromSSR && actionsToSkip && actionsToSkip.has(effectActionType) ? 1 : 0;
                const payload$ = action$.pipe(filter(({ type }) => type === effectActionType), skip(skipCount), map(({ payload }) => payload));
                return effect(payload$);
            }));
        };
    }
    combineReducers() {
        const reducerKeys = Reflect.getMetadata(REDUCER_DECORATOR_SYMBOL, this.constructor) || [];
        const reducers = reducerKeys.reduce((acc, property) => {
            acc[property] = this[property].bind(this);
            return acc;
        }, {});
        const immerReducerKeys = Reflect.getMetadata(IMMER_REDUCER_DECORATOR_SYMBOL, this.constructor) || [];
        this._actionKeys.push(...reducerKeys, ...immerReducerKeys);
        const immerReducers = immerReducerKeys.reduce((acc, property) => {
            acc[property] = this[property].bind(this);
            return acc;
        }, {});
        return (prevState, action) => {
            const { type } = action;
            if (typeof type === 'string') {
                if (reducers[type]) {
                    return reducers[type](prevState, action.payload);
                }
                else if (immerReducers[type]) {
                    return produce(prevState, (draft) => immerReducers[type](draft, action.payload));
                }
            }
            return prevState;
        };
    }
    combineDefineActions() {
        this._defineActionKeys =
            Reflect.getMetadata(DEFINE_ACTION_DECORATOR_SYMBOL, this.constructor) || [];
        this._actionKeys.push(...this._defineActionKeys);
        for (const actionType of this._defineActionKeys) {
            ;
            this[actionType] = this.action$.pipe(filter(({ type }) => type === actionType), map(({ payload }) => payload), publish(), refCount());
        }
    }
}
