import { from, race, timer, throwError, Subject, noop, Observable } from 'rxjs';
import { flatMap, bufferCount, take, filter, tap } from 'rxjs/operators';
import { rootInjector } from '@asuka/di';
import { SSRSymbol } from './constants';
import { SKIP_SYMBOL } from './ssr-effect';
import { TERMINATE_ACTION } from './terminate';
import { SSRStateCacheInstance } from './ssr-states';
import { oneShotCache } from './ssr-oneshot-cache';
const skipFn = () => SKIP_SYMBOL;
/**
 * Run all @SSREffect decorated effects of given modules and extract latest states.
 * `cleanup` function returned must be called before end of responding
 *
 * @param ctx request context, which will be passed to payloadGetter in SSREffect decorator param
 * @param modules used ayanami modules
 * @param uuid the same uuid would reuse the same state which was created before
 * @param timeout seconds to wait before all effects stream out TERMINATE_ACTION
 * @returns object contains ayanami state and cleanup function
 */
export const runSSREffects = (ctx, modules, sharedCtx, timeout = 3) => {
    const stateToSerialize = {};
    return modules.length === 0
        ? Promise.resolve(stateToSerialize)
        : race(from(modules).pipe(flatMap((constructor) => {
            return new Observable((observer) => {
                let cleanup = noop;
                const metas = Reflect.getMetadata(SSRSymbol, constructor.prototype) || [];
                let ayanamiState;
                let moduleName;
                const middleware = (effect$) => effect$.pipe(tap({
                    error: (e) => {
                        observer.error(e);
                    },
                }));
                if (sharedCtx) {
                    if (SSRStateCacheInstance.has(sharedCtx, constructor)) {
                        ayanamiState = SSRStateCacheInstance.get(sharedCtx, constructor);
                        moduleName = constructor.prototype.moduleName;
                    }
                    else {
                        const ayanamiInstance = rootInjector.resolveAndInstantiate(constructor);
                        moduleName = ayanamiInstance.moduleName;
                        ayanamiState = ayanamiInstance.createState(middleware);
                        SSRStateCacheInstance.set(sharedCtx, constructor, ayanamiState);
                    }
                }
                else {
                    const ayanamiInstance = rootInjector.resolveAndInstantiate(constructor);
                    moduleName = ayanamiInstance.moduleName;
                    ayanamiState = ayanamiInstance.createState(middleware);
                    oneShotCache.store(ctx, constructor, ayanamiState);
                }
                let effectsCount = metas.length;
                let disposeFn = noop;
                cleanup = sharedCtx
                    ? () => disposeFn()
                    : () => {
                        ayanamiState.unsubscribe();
                    };
                async function runEffects() {
                    await Promise.all(metas.map(async (meta) => {
                        if (meta.middleware) {
                            const param = await meta.middleware(ctx, skipFn);
                            if (param !== SKIP_SYMBOL) {
                                ayanamiState.dispatch({
                                    type: meta.action,
                                    payload: param,
                                    state: ayanamiState,
                                });
                            }
                            else {
                                effectsCount -= 1;
                            }
                        }
                        else {
                            ayanamiState.dispatch({
                                type: meta.action,
                                payload: undefined,
                                state: ayanamiState,
                            });
                        }
                    }));
                    if (effectsCount > 0) {
                        const action$ = new Subject();
                        disposeFn = ayanamiState.subscribeAction((action) => {
                            action$.next(action);
                        });
                        await action$
                            .pipe(filter((act) => act.type === TERMINATE_ACTION.type), bufferCount(effectsCount), take(1))
                            .toPromise();
                        const state = ayanamiState.getState();
                        stateToSerialize[moduleName] = state;
                    }
                }
                runEffects()
                    .then(() => {
                    observer.next(stateToSerialize);
                    observer.complete();
                })
                    .catch((e) => {
                    observer.error(e);
                });
                return cleanup;
            });
        })), timer(timeout * 1000).pipe(flatMap(() => throwError(new Error('Terminate timeout'))))).toPromise();
};
